#!/usr/bin/env python
# -*- coding: utf-8 -*-

from math import log10, floor
import warnings
import numpy as np
from optuna.exceptions import ExperimentalWarning
import pandas as pd
from sklearn.exceptions import ConvergenceWarning
from sklearn.model_selection import train_test_split
from covsirphy.util.argument import find_args
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.term import Term
from covsirphy.regression.reg_pred_actual_plot import _PredActualPlot


class _RegressorBase(Term):
    """
    Basic class to predict parameter values of ODE models.

    Args:
        X (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                (int/float): indicators
        y (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                (int/float) target values
        delay_values (list[int]): list of delay period [days]
        kwargs: keyword arguments of sklearn.model_selection.train_test_split(test_size=0.2, random_state=0)

    Note:
        If @seed is included in kwargs, this will be converted to @random_state.
    """
    warnings.simplefilter("ignore", category=ConvergenceWarning)
    warnings.simplefilter("ignore", category=ExperimentalWarning)
    # Description of regressor
    DESC = ""

    def __init__(self, X, y, delay_values, **kwargs):
        # Validate values
        self._ensure_dataframe(X, name="X", time_index=True)
        self._ensure_dataframe(y, name="y", time_index=True)
        self._delay_values = [self._ensure_natural_int(v) for v in self._ensure_list(delay_values)]
        # Set training/test dataset
        splitted_all = self._split(X, y, self._delay_values, **kwargs)
        self._X_train, self._X_test, self._y_train, self._y_test, self._X_target = splitted_all
        # Regression model
        self._pipeline = None
        self._param = {}
        # Perform fitting
        self._fit()

    @staticmethod
    def _split(X, y, delay_values, **kwargs):
        """
        Add delayed indicator values to X and split X and y to train/test data.

        Args:
            X (pandas.DataFrame): indicators with time index
            y (pandas.DataFrame): target values with time index
            delay_values (list[int]): list of delay period [days]
            kwargs: keyword arguments of sklearn.model_selection.train_test_split()

        Returns:
            tuple(pandas.DataFrame): datasets with time index
                - X_train
                - X_test
                - y_train
                - y_test
                - X_target

        Note:
            If @seed is included in kwargs, this will be converted to @random_state.

        Note:
            default values regarding sklearn.model_selection.train_test_split() are
            test_size=0.2, random_state=0, shuffle=False.
        """
        split_kwargs = {"test_size": 0.2, "random_state": 0, "shuffle": False}
        split_kwargs.update(kwargs)
        split_kwargs["random_state"] = split_kwargs.get("seed", split_kwargs["random_state"])
        split_kwargs = find_args(train_test_split, **split_kwargs)
        # Add delayed indicator values to X
        X_delayed = X.copy()
        for delay in delay_values:
            X_delayed = X_delayed.join(X.shift(delay, freq="D"), how="outer", rsuffix=f"_{delay}")
        X_delayed = X_delayed.ffill()
        # Training/test data
        df = X_delayed.join(y, how="inner").dropna().drop_duplicates()
        X_arranged = df.loc[:, X_delayed.columns]
        y_arranged = df.loc[:, y.columns]
        splitted = train_test_split(X_arranged, y_arranged, **split_kwargs)
        # X_target
        X_target = X_delayed.loc[X_delayed.index > y.index.max()]
        return (*splitted, X_target)

    def _fit(self):
        """
        Fit regression model with training dataset, update self._pipeline and self._param.
        This method should be overwritten by child classes.
        """
        raise NotImplementedError

    def score_train(self, metric):
        """
        Calculate score with training dataset.

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()

        Returns:
            float: evaluation score
        """
        pred_train = pd.DataFrame(self._pipeline.predict(self._X_train), columns=self._y_train.columns)
        return Evaluator(pred_train, self._y_train, how="all").score(metric=metric)

    def score_test(self, metric):
        """
        Calculate score with test dataset.

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()

        Returns:
            float: evaluation score
        """
        pred_test = pd.DataFrame(self._pipeline.predict(self._X_test), columns=self._y_test.columns)
        return Evaluator(pred_test, self._y_test, how="all").score(metric=metric)

    def to_dict(self, metric):
        """
        Calculate score with training/test dataset.

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()

        Returns:
            dict(str, object)
        """
        try:
            param_dict = {
                **self._pipeline.best_params_,
                "intercept": pd.DataFrame(),
                "coef": pd.DataFrame(),
            }
        except AttributeError:
            param_dict = {}
        param_dict.update(self._param)
        return {
            **param_dict,
            "score_metric": metric,
            "score_train": self.score_train(metric=metric),
            "score_test": self.score_test(metric=metric),
            "delay": self._delay_values,
            "dataset": {
                "X_train": self._X_train,
                "y_train": self._y_train,
                "X_test": self._X_test,
                "y_test": self._y_test,
                "X_target": self._X_target,
            }
        }

    def predict(self):
        """
        Predict parameter values (via y) with self._pipeline and X_target.

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): future dates
                Columns
                    (float): parameter values (4 digits)
        """
        # Predict parameter values
        predicted = self._pipeline.predict(self._X_target)
        df = pd.DataFrame(predicted, index=self._X_target.index, columns=self._y_train.columns)
        # parameter values: 4 digits
        return df.applymap(lambda x: self._round(x, digits=4))

    def _round(self, value, digits):
        """
        Round off the value to @digits significant digits.

        Args:
            value (float): target value
            digists (int): significant digits

        Return:
            float: rounded value
        """
        return 0 if not value else np.around(value, digits - int(floor(log10(abs(value)))) - 1)

    def _float2str(self, value, to_percentage):
        """
        Convert a float to a string.

        Args:
            value (float): target value
            to_percentage (bool): whether show the value as a percentage or not

        Returns:
            str: the string
        """
        return f"{value:.2%}" if to_percentage else f"{value:.3g}"

    def pred_actual_plot(self, metric, filename=None):
        """
        Create a scatter plot (predicted vs. actual parameter values).

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()
            fileaname (str): filename of the figure or None (display)
        """
        TITLE = f"Predicted vs. actual parameter values\n{self.DESC}"
        PRED, ACTUAL = "Predicted values", "Actual values"
        # Scores
        train_score = self.score_train(metric=metric)
        test_score = self.score_test(metric=metric)
        # Legend
        to_percentage = metric.upper() == "MAPE"
        train_score_str = self._float2str(train_score, to_percentage=to_percentage)
        test_score_str = self._float2str(test_score, to_percentage=to_percentage)
        train_title = f"Training data (n={len(self._X_train)}, {metric}={train_score_str})"
        test_title = f"Test data (n={len(self._X_test)}, {metric}={test_score_str})"
        # Predicted & training
        pred_train = pd.DataFrame(self._pipeline.predict(self._X_train), columns=self._y_train.columns)
        pred_train["subset"] = train_title
        # Predicted & test
        pred_test = pd.DataFrame(self._pipeline.predict(self._X_test), columns=self._y_test.columns)
        pred_test["subset"] = test_title
        # Actual & training
        act_train = self._y_train.copy()
        act_train["subset"] = train_title
        # Actual & test
        act_test = self._y_train.copy()
        act_test["subset"] = test_title
        # Combine data: index=reset, columns=parameter/subset/Predicted/Actual
        df = pd.concat([pred_train, pred_test], ignore_index=True)
        df = df.melt(id_vars=["subset"], var_name="parameter", value_name=PRED)
        act_df = pd.concat([act_train, act_test], ignore_index=True)
        act_df = act_df.melt(id_vars=["subset"], var_name="parameter", value_name=ACTUAL)
        df.loc[:, ACTUAL] = act_df.loc[:, ACTUAL]
        # Plotting
        with _PredActualPlot(filename=filename) as pa:
            pa.plot(df, x=ACTUAL, y=PRED)
            pa.title = TITLE



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
import ruptures as rpt
from covsirphy.util.error import deprecate
from covsirphy.util.term import Term
from covsirphy.trend.sr_change import _SRChange


class TrendDetector(Term):
    """
    Interface for trend analysis (change point analysis).

    Args:
        data (pandas.DataFrame): data to analyse
            Index:
                reset index
            Column:
                - Date(pd.Timestamp): Observation date
                - Confirmed(int): the number of confirmed cases
                - Infected(int): the number of currently infected cases
                - Fatal(int): the number of fatal cases
                - Recovered (int): the number of recovered cases
                - Susceptible(int): the number of susceptible cases
        area (str): area name (used in the figure title)
        min_size (int): minimum value of phase length [days], over 2

    Note:
        "Phase" means a sequential dates in which the parameters of SIR-derived models are fixed.
        "Change points" means the dates when trend was changed.
        "Change points" is the same as the start dates of phases except for the 0th phase.
    """

    def __init__(self, data, area="Selected area", min_size=7):
        self._ensure_dataframe(data, name="data", columns=self.SUB_COLUMNS)
        # Index: Date, Columns: the number cases
        self._record_df = data.groupby(self.DATE).last()
        # Minimum size of phases
        self._min_size = self._ensure_int_range(min_size, name="min_size", value_range=(3, None))
        if len(self._record_df) < self._min_size * 2:
            raise ValueError(f"More than {min_size * 2} records must be included because @min_size is {min_size}.")
        # The first/last date
        self._first_point = self._record_df.index.min()
        self._last_point = self._record_df.index.max()
        # Area name
        self._area = area
        # Change points: list[pandas.Timestamp]
        self._points = []

    def reset(self):
        """
        Reset the phase setting with the end dates of the records.

        Returns:
            covsirphy.TrendDetector: self
        """
        self._points = []
        return self

    def dates(self):
        """
        Return the list of start dates and end dates.

        Returns:
            tuple(list[str], list[str]): list of start dates and end dates
        """
        points = self._points[:]
        # Start dates
        start_dates = [date.strftime(self.DATE_FORMAT) for date in [self._first_point, *points]]
        # End dates
        end_dates = [self.yesterday(date.strftime(self.DATE_FORMAT)) for date in points]
        end_dates.append(self._last_point.strftime(self.DATE_FORMAT))
        return (start_dates, end_dates)

    def summary(self, metric=None, metrics="MSE"):
        """
        Summarize the phases with a dataframe.

        Args:
            metric (str or None): metric name or None (use @metrics)
            metrics (str): alias of @metric

        Returns:
            pandas.Dataframe:
                Index
                    (str): phase names
                Columns
                    - Start (str): star dates
                    - End (str): end dates
                    - Duration (int): phase duration
                    - {metric}_S-R (float): scores on S-R plane with the selected metric

        Note:
            Please refer to covsirphy.Evaluator.score() for metric names
        """
        metric = metric or metrics
        # Phase duration
        start_dates, end_dates = self.dates()
        duration_list = [self.steps(start, end, tau=1440) for (start, end) in zip(start_dates, end_dates)]
        # Scores in S-R plane
        scores = _SRChange(sr_df=self._record_df).score(change_points=self._points, metric=metric)
        return pd.DataFrame(
            {
                self.START: start_dates,
                self.END: end_dates,
                "Duration": duration_list,
                f"{metric}_S-R": scores,
            },
            index=[self.num2str(num) for num in range(len(self._points) + 1)]
        )

    def sr(self, algo="Binseg-normal", **kwargs):
        """
        Perform S-R trend analysis.

        Args:
            algo (str): detection algorithms and models
            kwargs: the other arguments of algorithm classes (ruptures.Pelt, .Binseg, BottomUp)

        Raises:
            UnExpectedValueError: un-expected value was applied as algorithm name

        Returns:
            covsirphy.TrendDetector: self

        Note:
            Candidates of @algo are "Pelt-rbf", "Binseg-rbf", "Binseg-normal", "BottomUp-rbf", "BottomUp-normal".
            Please refer to documentation of ruptures package.
            https://centre-borelli.github.io/ruptures-docs/
        """
        # Set algorithm class
        algo_kwargs = {"jump": 1, "min_size": self._min_size}
        algo_kwargs.update(kwargs)
        algo_dict = {
            "Pelt-rbf": (rpt.Pelt, {"model": "rbf"}),
            "Binseg-rbf": (rpt.Binseg, {"model": "rbf"}),
            "Binseg-normal": (rpt.Binseg, {"model": "normal"}),
            "BottomUp-rbf": (rpt.BottomUp, {"model": "rbf"}),
            "BottomUp-normal": (rpt.BottomUp, {"model": "normal"}),
        }
        self._ensure_selectable(target=algo, candidates=list(algo_dict.keys()), name="algo")
        algo_kwargs.update(algo_dict[algo][1])
        algorithm = algo_dict[algo][0](**algo_kwargs)
        # Run trend analysis
        finder = _SRChange(sr_df=self._record_df)
        points = finder.run(algorithm=algorithm, **algo_kwargs)
        self._points = sorted(set(self._points) | set(points))
        return self

    def show(self, **kwargs):
        """
        Show the trend on S-R plane.

        Args:
            kwargs: keyword arguments of covsirphy.trend_plot()
        """
        finder = _SRChange(sr_df=self._record_df)
        finder.show(self._points, self._area, **kwargs)


class Trend(TrendDetector):
    """
    Deprecated. Please use TrendDetector class.
    """
    @deprecate("covsirphy.Trend", new="covsirphy.TrendDetector", version="2.17.0-zeta")
    def __init__(self, data, area="Selected area", min_size=5):
        super().__init__(data, area=area, min_size=min_size)


class ChangeFinder(TrendDetector):
    """
    Deprecated. Please use TrendDetector class.
    """
    @deprecate("covsirphy.ChangeFinder", new="covsirphy.TrendDetector", version="2.17.0-zeta")
    def __init__(self, data, area="Selected area", min_size=5):
        super().__init__(data, area=area, min_size=min_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from lightgbm import LGBMRegressor
from sklearn.decomposition import PCA
from sklearn.model_selection import GridSearchCV, TimeSeriesSplit
from sklearn.multioutput import MultiOutputRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from covsirphy.regression.regbase import _RegressorBase
from covsirphy.regression.reg_rate_converter import _RateConverter


class _ParamLightGBMRegressor(_RegressorBase):
    """
    Predict parameter values of ODE models with light gradient boosting machine regressor.

    Args:
        X (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                (int/float): indicators
        y (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                (int/float) target values
        delay_values (list[int]): list of delay period [days]
        kwargs: keyword arguments of sklearn.model_selection.train_test_split()

    Note:
        If @seed is included in kwargs, this will be converted to @random_state.

    Note:
        default values regarding sklearn.model_selection.train_test_split() are
        test_size=0.2, random_state=0, shuffle=False.
    """
    # Description of regressor
    DESC = "Indicators -> Parameters with Light Gradient Boosting Machine Regressor"

    def _fit(self):
        """
        Fit regression model with training dataset, update self._pipeline and self._param.
        """
        # Paramters of the steps
        param_grid = {
            "converter__to_convert": [True, False],
            "pca__n_components": [0.3, 0.5, 0.7, 0.9],
            "regressor__estimator__max_depth": list(range(1, 5)),
        }
        # Fit with pipeline
        steps = [
            ("converter", _RateConverter()),
            ("scaler", MinMaxScaler()),
            ("pca", PCA(random_state=0)),
            ("regressor", MultiOutputRegressor(LGBMRegressor(n_estimators=200, random_state=0))),
        ]
        tscv = TimeSeriesSplit(n_splits=5).split(self._X_train)
        pipeline = GridSearchCV(Pipeline(steps=steps), param_grid, n_jobs=-1, cv=tscv)
        pipeline.fit(self._X_train, self._y_train)
        # Update regressor
        self._pipeline = pipeline
        # Update param
        self._param.update(**{k: type(v) for (k, v) in steps})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import pandas as pd
from covsirphy.util.error import SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class OxCGRTData(CleaningBase):
    """
    Data cleaning of OxCGRT dataset.

    Args:
        filename (str or None): CSV filename of the dataset
        data (pandas.DataFrame or None):
            Index
                reset index
            Columns
                - Date: Observation date
                - ISO3: ISO 3166-1 alpha-3, like JPN
                - Country: country/region name
                - School_closing
                - Workplace_closing
                - Cancel_events
                - Gatherings_restrictions
                - Transport_closing
                - Stay_home_restrictions
                - Internal_movement_restrictions
                - International_movement_restrictions
                - Information_campaigns
                - Testing_policy
                - Contact_tracing
                - Stringency_index
        citation (str or None): citation or None (empty)

    Note:
        Either @filename (high priority) or @data must be specified.

    Note:
        Policy indices (Overall etc.) are from README.md and documentation/index_methodology.md in
        https://github.com/OxCGRT/covid-policy-tracker/
    """
    OXCGRT_VARS = [
        "School_closing",
        "Workplace_closing",
        "Cancel_events",
        "Gatherings_restrictions",
        "Transport_closing",
        "Stay_home_restrictions",
        "Internal_movement_restrictions",
        "International_movement_restrictions",
        "Information_campaigns",
        "Testing_policy",
        "Contact_tracing",
        "Stringency_index"
    ]
    # Columns of self._raw and self._clean_df
    RAW_COLS = [CleaningBase.DATE, CleaningBase.ISO3, CleaningBase.COUNTRY, *OXCGRT_VARS]
    # Columns of self.cleaned()
    CLEANED_COLS = RAW_COLS[:]
    # Columns of self.subset()
    SUBSET_COLS = [CleaningBase.DATE, *OXCGRT_VARS]
    # Indicators except for Stringency index
    OXCGRT_VARS_INDICATORS = [v for v in OXCGRT_VARS if v != "Stringency_index"]
    # Deprecated
    OXCGRT_VARIABLES_RAW = [v.lower() for v in OXCGRT_VARS]
    OXCGRT_COLS = RAW_COLS[:]
    OXCGRT_COLS_WITHOUT_COUNTRY = SUBSET_COLS[:]

    def __init__(self, filename=None, data=None, citation=None):
        # Raw data
        self._raw = self._parse_raw(filename, data, self.RAW_COLS)
        # Data cleaning
        self._cleaned_df = pd.DataFrame(columns=self.RAW_COLS) if self._raw.empty else self._cleaning()
        # Citation
        self._citation = citation or ""
        # Directory that save the file
        if filename is None:
            self._dirpath = Path("input")
        else:
            Path(filename).parent.mkdir(exist_ok=True, parents=True)
            self._dirpath = Path(filename).resolve().parent

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - ISO3 (str): ISO 3166-1 alpha-3, like JPN
                    - Country (pandas.Category): country/region name
                    - School_closing
                    - Workplace_closing
                    - Cancel_events
                    - Gatherings_restrictions
                    - Transport_closing
                    - Stay_home_restrictions
                    - Internal_movement_restrictions
                    - International_movement_restrictions
                    - Information_campaigns
                    - Testing_policy
                    - Contact_tracing
                    - Stringency_index
        """
        df = self._raw.copy()
        # Prepare data for Greenland
        grl_df = df.loc[df[self.COUNTRY] == "Denmark"].copy()
        grl_df.loc[:, [self.ISO3, self.COUNTRY]] = ["GRL", "Greenland"]
        df = pd.concat([df, grl_df], sort=True, ignore_index=True)
        # Confirm the expected columns are in raw data
        self._ensure_dataframe(df, name="the raw data", columns=self.CLEANED_COLS)
        # Read date records
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Confirm float type
        for col in self.OXCGRT_VARS:
            df[col] = pd.to_numeric(df[col], errors="coerce").ffill()
        # Update data types to reduce memory
        cat_cols = [self.ISO3, self.COUNTRY]
        df[cat_cols] = df[cat_cols].astype("category")
        return df.loc[:, self.CLEANED_COLS]

    def subset(self, country, **kwargs):
        """
        Create a subset for a country.

        Args:
            country (str): country name or ISO 3166-1 alpha-3, like JPN
            kwargs: the other arguments will be ignored in the latest version.

        Raises:
            covsirphy.SubsetNotFoundError: no records were found

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - School_closing
                    - Workplace_closing
                    - Cancel_events
                    - Gatherings_restrictions
                    - Transport_closing
                    - Stay_home_restrictions
                    - Internal_movement_restrictions
                    - International_movement_restrictions
                    - Information_campaigns
                    - Testing_policy
                    - Contact_tracing
                    - Stringency_index
        """
        country_arg = country
        country = self.ensure_country_name(country)
        try:
            df = super().subset(country=country)
        except SubsetNotFoundError:
            raise SubsetNotFoundError(country=country_arg, country_alias=country) from None
        df = df.groupby(self.DATE).last().reset_index()
        return df.loc[:, self.SUBSET_COLS]

    def total(self):
        """
        This method is not defined for OxCGRTData class.
        """
        raise NotImplementedError

    def map(self, country=None, variable="Stringency_index", date=None, **kwargs):
        """
        Create global colored map to show the values.

        Args:
            country (None): always None
            variable (str): variable name to show
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()

        Raises:
            NotImplementedError: @country was specified
        """
        if country is not None:
            raise NotImplementedError("@country cannot be specified, always None.")
        # Date
        date_str = date or self.cleaned()[self.DATE].max().strftime(self.DATE_FORMAT)
        country_str = country or "Global"
        title = f"{country_str}: {variable.lower().replace('_', ' ')} on {date_str}"
        # Global map
        return self._colored_map_global(variable=variable, title=title, date=date, logscale=False, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import warnings
import covid19dh
import pandas as pd
from covsirphy.util.term import Term
from covsirphy.cleaning.cbase import CleaningBase
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.cleaning.oxcgrt import OxCGRTData
from covsirphy.cleaning.population import PopulationData
from covsirphy.cleaning.pcr_data import PCRData


class COVID19DataHub(Term):
    """
    Load datasets retrieved from COVID-19 Data Hub.
    https://covid19datahub.io/

    Args:
        filename (str): CSV filename to save records
    """
    # Citation
    CITATION = '(Secondary source)' \
        ' Guidotti, E., Ardia, D., (2020), "COVID-19 Data Hub",' \
        ' Journal of Open Source Software 5(51):2376, doi: 10.21105/joss.02376.'
    # Name conversion list of columns
    _OXCGRT_COLS_RAW = [
        "school_closing",
        "workplace_closing",
        "cancel_events",
        "gatherings_restrictions",
        "transport_closing",
        "stay_home_restrictions",
        "internal_movement_restrictions",
        "international_movement_restrictions",
        "information_campaigns",
        "testing_policy",
        "contact_tracing",
        "stringency_index"
    ]
    _COL_DICT = {
        "date": Term.DATE,
        "administrative_area_level_1": Term.COUNTRY,
        "administrative_area_level_2": Term.PROVINCE,
        "tests": Term.TESTS,
        "confirmed": Term.C,
        "deaths": Term.F,
        "recovered": Term.R,
        "population": Term.N,
        "iso_alpha_3": Term.ISO3,
        **{v: v.capitalize() for v in _OXCGRT_COLS_RAW}
    }
    # Class objects of datasets
    OBJ_DICT = {
        "jhu": JHUData,
        "population": PopulationData,
        "oxcgrt": OxCGRTData,
        "pcr": PCRData,
    }

    def __init__(self, filename):
        try:
            self.filepath = Path(filename)
        except TypeError:
            raise TypeError(f"@filename should be a path-like object, but {filename} was applied.")
        self.filepath.parent.mkdir(exist_ok=True, parents=True)
        self.primary_list = None
        self._loaded_df = None

    def load(self, name="jhu", force=True, verbose=1):
        """
        Load the datasets of COVID-19 Data Hub and create dataset object.

        Args:
            name (str): name of dataset, "jhu", "population", "oxcgrt" or "pcr"
            force (bool): if True, always download the dataset from the server
            verbose (int): level of verbosity

        Returns:
            covsirphy.CleaningBase: the dataset

        Note:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
            Citation of COVID-19 Data Hub will be set as JHUData.citation etc.
        """
        if name not in self.OBJ_DICT:
            raise KeyError(
                f"@name must be {', '.join(list(self.OBJ_DICT.keys()))}, but {name} was applied.")
        # Get all data
        if self._loaded_df is None:
            self._loaded_df = self._load(force=force, verbose=verbose)
        return self.OBJ_DICT[name](data=self._loaded_df, citation=self.CITATION)

    def _load(self, force, verbose):
        """
        Load the datasets of COVID-19 Data Hub.

        Args:
            force (bool): if True, always download the dataset from the server
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame: as the same as COVID19DataHub._preprocessing()

        Note:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
        """
        # Use local CSV file
        if not force and self.filepath.exists():
            df = CleaningBase.load(
                self.filepath,
                dtype={
                    self.PROVINCE: "object", "Province/State": "object",
                    "key": "object", "key_alpha_2": "object",
                })
            if set(self._COL_DICT.values()).issubset(df.columns):
                return df
        # Download dataset from server
        raw_df = self._retrieve(verbose=verbose)
        raw_df.to_csv(self.filepath, index=False)
        return raw_df

    def _retrieve(self, verbose=1):
        """
        Retrieve datasets from COVID-19 Data Hub.
        Level 1 (country) and level2 (province) will be used and combined to a dataframe.

        Args:
            verbose (int): level of verbosity

        Note:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
        """
        # Download datasets
        if verbose:
            print("Retrieving datasets from COVID-19 Data Hub https://covid19datahub.io/")
        c_df, p_df, self.primary_list = self._download()
        # Merge the datasets
        df = pd.concat([c_df, p_df], axis=0, ignore_index=True)
        # Perform pre-processing
        df = self._preprocessing(df)
        # Show citation list
        if verbose:
            if isinstance(verbose, int) and verbose >= 2:
                print("\nDetailed citaition list:")
                print(self.primary_list)
            else:
                print("\tPlease set verbose=2 to see the detailed citation list.")
        return df

    def _download(self):
        """
        Retrieve dataset and citation list from COVID-19 Data Hub.

        Returns:
            tuple:
                pandas.DataFrame: dataset at country level
                pandas.DataFrame: dataset at province level
                str: the list of primary sources

        Note:
            For some countries, province-level data is included.
        """
        warnings.simplefilter("ignore", ResourceWarning)
        levels = [f"administrative_area_level_{i}" for i in range(1, 4)]
        # Level 1 (country/region)
        c_raw, c_cite = covid19dh.covid19(country=None, level=1, verbose=False, raw=True)
        c_df = c_raw.groupby(levels[0]).ffill().fillna(0)
        for num in range(3):
            c_df.loc[:, levels[num]] = c_raw[levels[num]]
        # Level 2 (province/state)
        p_raw, p_cite = covid19dh.covid19(country=None, level=2, verbose=False, raw=True)
        p_df = p_raw.groupby(levels[:2]).ffill().fillna(0)
        for num in range(3):
            p_df.loc[:, levels[num]] = p_raw[levels[num]]
        # Citation
        cite = pd.concat([c_cite, p_cite], axis=0, ignore_index=True)
        cite = cite.loc[:, ["title", "year", "url"]]
        cite = cite.sort_values(["year", "url"], ascending=[False, True])
        cite.drop_duplicates(subset="title", inplace=True)
        series = cite.apply(lambda x: f"{x[0]} ({x[1]}), {x[2]}", axis=1)
        return (c_df, p_df, "\n".join(series.tolist()))

    def _preprocessing(self, raw):
        """
        Perform pre-processing with the raw dataset.

        Args:
            raw (pandas.DataFrame):
                Index
                    reset index
                Columns
                    Refer to COVID-19 Data Hub homepage.
                    https://covid19datahub.io/articles/doc/data.html

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date: observation date
                    - Country: country/region name
                    - Province: province/prefecture/state name
                    - Confirmed: the number of confirmed cases
                    - Infected: the number of currently infected cases
                    - Fatal: the number of fatal cases
                    - Recovered: the number of recovered cases
                    - School_closing
                    - Workplace_closing
                    - Cancel_events
                    - Gatherings_restrictions
                    - Transport_closing
                    - Stay_home_restrictions
                    - Internal_movement_restrictions
                    - International_movement_restrictions
                    - Information_campaigns
                    - Testing_policy
                    - Contact_tracing
                    - Stringency_index

        Note:
            Data types are not confirmed.
        """
        df = raw.copy()
        # Replace column names
        df = df.rename(columns=self._COL_DICT)
        self._ensure_dataframe(df, columns=list(self._COL_DICT.values()))
        # Country
        df[self.COUNTRY] = df[self.COUNTRY].replace(
            {
                # COD
                "Congo, the Democratic Republic of the": "Democratic Republic of the Congo",
                # COG
                "Congo": "Republic of the Congo",
                # KOR
                "Korea, South": "South Korea",
            }
        )
        # Set 'Others' as the country name of cruise ships
        ships = ["Diamond Princess", "Costa Atlantica", "Grand Princess", "MS Zaandam"]
        for ship in ships:
            df.loc[df[self.COUNTRY] == ship, [self.COUNTRY, self.PROVINCE]] = [self.OTHERS, ship]
        return df

    @property
    def primary(self):
        """
        str: the list of primary sources.
        """
        return self.primary_list or self._download()[-1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from sklearn.linear_model import MultiTaskElasticNetCV
from sklearn.model_selection import TimeSeriesSplit
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from covsirphy.regression.regbase import _RegressorBase
from covsirphy.regression.reg_rate_converter import _RateConverter


class _ParamElasticNetRegressor(_RegressorBase):
    """
    Predict parameter values of ODE models with Elastic Net regression.

    Args:
        X (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                (int/float): indicators
        y (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                (int/float) target values
        delay_values (list[int]): list of delay period [days]
        kwargs: keyword arguments of sklearn.model_selection.train_test_split()

    Note:
        If @seed is included in kwargs, this will be converted to @random_state.

    Note:
        default values regarding sklearn.model_selection.train_test_split() are
        test_size=0.2, random_state=0, shuffle=False.
    """
    # Description of regressor
    DESC = "Indicators -> Parameters with Elastic Net"

    def _fit(self):
        """
        Fit regression model with training dataset, update self._pipeline and self._param.
        """
        # Model for Elastic Net regression
        tscv = TimeSeriesSplit(n_splits=5).split(self._X_train)
        cv = MultiTaskElasticNetCV(
            alphas=[0, 0.001, 0.01, 0.1, 1, 10, 100, 1000],
            l1_ratio=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            cv=tscv, n_jobs=-1
        )
        # Fit with pipeline
        steps = [
            ("converter", _RateConverter(to_convert=False)),
            ("scaler", MinMaxScaler()),
            ("regressor", cv),
        ]
        pipeline = Pipeline(steps=steps)
        pipeline.fit(self._X_train, self._y_train)
        reg_output = pipeline.named_steps.regressor
        # Update regressor
        self._pipeline = pipeline
        # Intercept/coef
        intercept_df = pd.DataFrame(
            reg_output.coef_, index=self._y_train.columns, columns=self._X_train.columns)
        intercept_df.insert(0, "Intercept", None)
        intercept_df["Intercept"] = reg_output.intercept_
        # Update param
        param_dict = {
            **{k: type(v) for (k, v) in steps},
            "alpha": reg_output.alpha_,
            "l1_ratio": reg_output.l1_ratio_,
            "intercept": intercept_df,
            "coef": intercept_df,
        }
        self._param.update(param_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.error import UnExecutedError, deprecate
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.term import Term
from covsirphy.ode.mbase import ModelBase
from covsirphy.simulation.estimator import Estimator
from covsirphy.simulation.simulator import ODESimulator


class PhaseUnit(Term):
    """
    Save information of  a phase.

    Args:
        start_date (str): start date of the phase
        end_date (str): end date of the phase
        population (int): population value

    Examples:
        >>> unit1 = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        >>> unit2 = PhaseUnit("02Feb2020", "01Mar2020", 1000)
        >>> unit3 = PhaseUnit("02Mar2020", "01Apr2020", 1000)
        >>> unit4 = PhaseUnit("02Mar2020", "01Apr2020", 1000)
        >>> unit5 = PhaseUnit("01Jan2020", "01Apr2020", 1000)
        >>> str(unit1)
        'Phase (01Jan2020 - 01Feb2020)'
        >>> unit4 == unit4
        True
        >>> unit1 != unit2
        True
        >>> unit1 < unit2
        True
        >>> unit3 > unit1
        True
        >>> unit3 < unit4
        False
        >>> unit3 <= unit4
        True
        >>> unit1 < "02Feb2020"
        True
        >>> unit1 <= "01Feb2020"
        True
        >>> unit1 > "31Dec2019"
        True
        >>> unit1 >= "01Jan2020"
        True
        >>> sorted([unit3, unit1, unit2]) == [unit1, unit2, unit3]
        True
        >>> str(unit1 + unit2)
        'Phase (01Jan2020 - 01Mar2020)'
        >>> str(unit5 - unit1)
        'Phase (02Feb2020 - 01Apr2020)'
        >>> str(unit5 - unit4)
        'Phase (01Jan2020 - 01Mar2020)'
        >>> set([unit1, unit3, unit4]) == set([unit1, unit3])
        True
    """

    @deprecate("PhaseUnit", new="ODEHandler", version="2.19.1-zeta-fu1")
    def __init__(self, start_date, end_date, population):
        self._ensure_date_order(start_date, end_date, name="end_date")
        self._start_date = start_date
        self._end_date = end_date
        self._population = self._ensure_population(population)
        # Summary of information
        self.info_dict = {
            self.START: start_date,
            self.END: end_date,
            self.N: population,
            self.ODE: None,
            self.RT: None
        }
        self._ode_dict = {self.TAU: None}
        self.day_param_dict = {}
        self.est_dict = {
            **{metric: None for metric in Evaluator.metrics()},
            self.TRIALS: None,
            self.RUNTIME: None
        }
        # Init
        self._id_dict = None
        self._enabled = True
        self._model = None
        self._record_df = pd.DataFrame()
        self.y0_dict = {}
        self._estimator = None

    def __str__(self):
        if self._id_dict is None:
            header = "Phase"
        else:
            id_str = ', '.join(list(self._id_dict.values()))
            header = f"{id_str:>4} phase"
        return f"{header} ({self._start_date} - {self._end_date})"

    def __hash__(self):
        return hash((self._start_date, self._end_date))

    def __eq__(self, other):
        if not isinstance(other, PhaseUnit):
            raise NotImplementedError
        return self._start_date == other.start_date and self._end_date == other.end_date

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        # self < other
        end = self._ensure_date(self._end_date)
        if isinstance(other, str):
            sta_other = self._ensure_date(other)
        elif isinstance(other, PhaseUnit):
            sta_other = self._ensure_date(other.start_date)
        else:
            raise NotImplementedError
        return end < sta_other

    def __le__(self, other):
        # self <= other
        end = self._ensure_date(self._end_date)
        if isinstance(other, str):
            sta_other = self._ensure_date(other)
        elif isinstance(other, PhaseUnit):
            if self.__eq__(other):
                return True
            sta_other = self._ensure_date(other.start_date)
        else:
            raise NotImplementedError
        return end <= sta_other

    def __gt__(self, other):
        # self > other
        if isinstance(other, PhaseUnit) and self.__eq__(other):
            return False
        return not self.__le__(other)

    def __ge__(self, other):
        # self >= other
        return not self.__lt__(other)

    def __add__(self, other):
        if self < other:
            return PhaseUnit(self._start_date, other.end_date, self._population)
        raise NotImplementedError

    def __iadd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        sta = self._ensure_date(self._start_date)
        end = self._ensure_date(self._end_date)
        sta_other = self._ensure_date(other.start_date)
        end_other = self._ensure_date(other.end_date)
        if sta < sta_other and end == end_other:
            end_date = self.yesterday(other.start_date)
            return PhaseUnit(self._start_date, end_date, self._population)
        if sta == sta_other and end > end_other:
            start_date = self.tomorrow(other.end_date)
            return PhaseUnit(start_date, self._end_date, self._population)

    def __isub__(self, other):
        return self.__sub__(other)

    def __contains__(self, date):
        sta = self._ensure_date(self._start_date)
        end = self._ensure_date(self._end_date)
        date = self._ensure_date(date)
        return sta <= date <= end

    @property
    def id_dict(self):
        """
        tuple(str): id_dict of the phase
        """
        return self._id_dict

    @ id_dict.setter
    def id_dict(self, value):
        self.set_id(value)

    def set_id(self, **kwargs):
        """
        Set identifiers.

        Args:
            id_dict (dict[str, str]): dictionary of identifiers

        Returns:
            covsirphy.PhaseUnit: self
        """
        if self._id_dict is not None:
            raise AttributeError("@id_dict cannot be overwritten.")
        self._id_dict = kwargs
        return self

    def del_id(self):
        """
        Delete identifers.

        Returns:
            covsirphy.PhaseUnit: self
        """
        self._id_dict = None
        return self

    def enable(self):
        """
        Enable the phase.

        Examples:
            >>> unit.enable
            >>> bool(unit)
            True
        """
        self._enabled = True

    def disable(self):
        """
        Disable the phase.

        Examples:
            >>> unit.disable
            >>> bool(unit)
            False
        """
        self._enabled = False

    def __bool__(self):
        return self._enabled

    @property
    def start_date(self):
        """
        str: start date
        """
        return self._start_date

    @property
    def end_date(self):
        """
        str: end date
        """
        return self._end_date

    @property
    def population(self):
        """
        str: population value
        """
        return self._population

    @property
    def tau(self):
        """
        int or None: tau value [min]
        """
        return self._ode_dict[self.TAU]

    @ tau.setter
    def tau(self, value):
        if self._ode_dict[self.TAU] is None:
            self._ode_dict[self.TAU] = self._ensure_tau(value)
            return
        raise AttributeError(
            f"PhaseUnit.tau is not None ({self._ode_dict[self.TAU]}) and cannot be changed.")

    @property
    def model(self):
        """
        covsirphy.ModelBase or None: model description
        """
        return self._model

    @property
    def estimator(self):
        """
        covsirphy.Estimator or None: estimator object
        """
        return self._estimator

    def to_dict(self):
        """
        Summarize phase information and return as a dictionary.

        Returns:
            dict:
                - Start: start date of the phase
                - End: end date of the phase
                - Population: population value of the start date
                - if available:
                    - ODE: model name
                    - Rt: (basic) reproduction number
                    - parameter values if available
                    - day parameter values if available
                    - tau: tau value [min]
                    - {metric name}: score of parameter estimation
                    - Trials: the number of trials in estimation
                    - Runtime: runtime of estimation
        """
        return {
            **self.info_dict,
            **self._ode_dict,
            **self.day_param_dict,
            **self.est_dict
        }

    def summary(self):
        """
        Summarize information.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Start: start date of the phase
                    - End: end date of the phase
                    - Population: population value of the start date
                    - if available:
                        - ODE (str): model name
                        - Rt (float): (basic) reproduction number
                        - rho etc. (float): parameter values if available
                        - tau (int): tau value [min]
                        - (int): day parameter values if available
                        - {metric name} (float): score of parameter estimation
                        - Trials (int): the number of trials in parameter estimation
                        - Runtime (str): runtime of parameter estimation
        """
        summary_dict = self.to_dict()
        df = pd.DataFrame.from_dict(summary_dict, orient="index").T
        return df.dropna(how="all", axis=1)

    def set_ode(self, model=None, tau=None, **kwargs):
        """
        Set ODE model, tau value and parameter values, if necessary.

        Args:
            model (covsirphy.ModelBase or None): ODE model
            tau (int or None): tau value [min], a divisor of 1440
            kwargs: keyword arguments of model parameters

        Returns:
            covsirphy.PhaseUnit: self
        """
        # Tau value
        tau = self._ensure_tau(tau) or self._ode_dict[self.TAU]
        # Model
        model = model or self._model
        if model is None:
            self._ode_dict[self.TAU] = tau
            return self
        self._model = self._ensure_subclass(model, ModelBase, name="model")
        self.info_dict[self.ODE] = model.NAME
        # Parameter values
        param_dict = self._ode_dict.copy()
        param_dict.update(kwargs)
        param_dict = {
            p: param_dict[p] if p in param_dict else None for p in model.PARAMETERS
        }
        self._ode_dict = param_dict.copy()
        self._ode_dict[self.TAU] = tau
        # Day parameters
        if None in param_dict.values():
            return self
        model_instance = model(population=self._population, **param_dict)
        self.info_dict[self.RT] = model_instance.calc_r0()
        # Reproduction number
        if tau is not None:
            self.day_param_dict = model_instance.calc_days_dict(tau)
        return self

    @property
    def record_df(self):
        """
        pandas.DataFrame: records of the phase

            Index
                reset index
            Columns
                - Date (pd.Timestamp): Observation date
                - Confirmed (int): the number of confirmed cases
                - Infected (int): the number of currently infected cases
                - Fatal (int): the number of fatal cases
                - Recovered (int): the number of recovered cases
                - Susceptible (int): the number of susceptible cases
        """
        return self._record_df

    @ record_df.setter
    def record_df(self, df):
        self._ensure_dataframe(df, name="df", columns=self.NLOC_COLUMNS)
        self._record_df = df.copy()
        self.set_y0(df)

    def _model_is_registered(self):
        """
        Ensure that model was set.

        Raises:
            NameError: ODE model is not registered
        """
        if self._model is None:
            raise UnExecutedError("PhaseUnit.set_ode(model)")

    def estimate(self, record_df=None, **kwargs):
        """
        Perform parameter estimation.

        Args:
            record_df (pandas.DataFrame or None)
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any other columns will be ignored
            **kwargs: keyword arguments of Estimator.run()

        Note:
            If @record_df is None, registered records will be used.
        """
        self._model_is_registered()
        # Records
        if record_df is None:
            record_df = self._record_df.copy()
        if record_df.empty:
            raise UnExecutedError(
                "PhaseUnit.record_df = ...", message="or specify @record_df argument")
        self._ensure_dataframe(record_df, name="record_df", columns=self.NLOC_COLUMNS)
        # Check dates
        sta = self._ensure_date(self.start_date)
        end = self._ensure_date(self.end_date)
        series = record_df[self.DATE]
        record_df = record_df.loc[(series >= sta) & (series <= end), :]
        # Parameter estimation of ODE model
        estimator = Estimator(
            record_df, self._model, self._population, **self._ode_dict, **kwargs)
        estimator.run(**kwargs)
        self._read_estimator(estimator, record_df)
        # Set estimator
        self._estimator = estimator

    def _read_estimator(self, estimator, record_df):
        """
        Read the result of parameter estimation and update the summary of phase.

        Args:
            estimator (covsirphy.Estimator): estimator which finished estimation
            record_df (pandas.DataFrame)
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any other columns will be ignored
        """
        # Reproduction number
        est_dict = estimator.to_dict()
        self.info_dict[self.RT] = est_dict.pop(self.RT)
        # Get parameter values and tau value
        ode_set = set([*self._model.PARAMETERS, self.TAU])
        ode_dict = {
            k: v for (k, v) in est_dict.items() if k in ode_set}
        self._ode_dict.update(ode_dict)
        # Other information of estimation
        other_dict = dict(est_dict.items() - ode_dict.items())
        self.est_dict.update(other_dict)
        self.est_dict = {k: v for (k, v) in self.est_dict.items() if v is not None}
        # Initial values
        self.set_y0(record_df=record_df)

    def set_y0(self, record_df):
        """
        Set initial values.

        Args:
            record_df (pandas.DataFrame)
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any other columns will be ignored
        """
        self._model_is_registered()
        df = record_df.loc[
            record_df[self.DATE] >= self._ensure_date(self.start_date), :]
        df = self._model.tau_free(df, self._population, tau=None)
        y0_dict = df.iloc[0, :].to_dict()
        self.y0_dict = {
            k: v for (k, v) in y0_dict.items() if k in set(self._model.VARIABLES)
        }

    def simulate(self, y0_dict=None):
        """
        Perform simulation with the set/estimated parameter values.

        Args:
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Variables of the model (int): Confirmed etc.

        Note:
            Simulation starts at the start date of the phase.
            Simulation end at the next date of the end date of the phase.
        """
        self._model_is_registered()
        # Initial values
        y0_dict = y0_dict or {}
        y0_dict.update(self.y0_dict)
        diff_set = set(self._model.VARIABLES) - y0_dict.keys()
        y0_dict.update({var: 0 for var in diff_set})
        # Conditions
        param_dict = self._ode_dict.copy()
        if None in param_dict.values():
            raise UnExecutedError("PhaseUnit.set_ode()")
        tau = param_dict.pop(self.TAU)
        last_date = self.tomorrow(self._end_date)
        # Simulation
        simulator = ODESimulator()
        simulator.add(
            model=self._model,
            step_n=self.steps(self._start_date, last_date, tau),
            population=self._population,
            param_dict=param_dict,
            y0_dict=y0_dict
        )
        # Dimensionalized values
        df = simulator.dim(tau=tau, start_date=self._start_date)
        df = self._model.restore(df)
        # Return day-level data
        df = df.set_index(self.DATE).resample("D").first()
        df = df.loc[df.index <= self._ensure_date(last_date), :]
        return df.reset_index().loc[:, self.NLOC_COLUMNS]
